MCMC <-
function(counts, nsurvive, mcmc=list( nsave=1000, nskip=1000,initc=1000,
	phi.init=1, phi.delta=0.1 ), method=1, print.iter=TRUE)
{
 ######################################################################
 # A function to run the MCMC procedure from Pei et al. 2012; 
 # for conditional p-value computation in Albertini melanoma data and
 # double over-dispersion model
 #
 # Gibbs on underlying frequencies in `theta'; Metropolis Hastings
 # on overdispersion parameter phi and latent states Z
 #
 # counts: a data set contains counts of V/J region types
 # nsurvive: expected number of cells past selection, per patient
 # mcmc:     list with MCMC conditions [nsave, nskip, initc, phi.init, phi.delta]
 # method: 1 CPP (condition on MT only),
 #         2 PPP (condition on all counts in posterior)
 ######################################################################

 # data elements
 wtfreqpt = counts[[1]]+counts[[2]]   ## Wild type frequencies
 mtfreqptSCDT = counts[[3]] ## Mutant single-cell frequencies
 mtfreqptMC = counts[[4]]   ## Mutant mass-culture frequencies
 types <- dimnames( counts[[1]] )[[1]]
 
 m = length(wtfreqpt[,1])  ## number of types
 n = length(wtfreqpt[1,])  ## number of patients
 
 # summary statistics
 wtfreq = apply(wtfreqpt,1,sum)
 mtfreq = apply(mtfreqptMC,1,sum)+apply(mtfreqptSCDT,1,sum)
 mtfreqSCDT = apply(mtfreqptSCDT,1,sum)

 ###########################################################
 # Run MCMC.

 phi <- mcmc$phi.init
 phi.delta <- mcmc$phi.delta  ## window in uniform update for phi
 nsave <- mcmc$nsave  ## store
 nskip <- mcmc$nskip  ## subsampel
 initc <- mcmc$initc  ## burn in

 acrate = rep(0,2)  # acceptance rates for zstate and phi updates

 #initialize other unknowns
 theta = (wtfreq+mtfreq)/(sum(wtfreq+mtfreq))
 zstate = matrix(NA, m, n )  
 muz = outer( theta, nsurvive, "*" )
 for( j in 1:n )
  {
    zstate[,j] = rpois( m, lambda=muz[,j] )  
  }
 zstate[mtfreqptMC>0 & zstate==0]=1  ## make sure state gives >0 prob to data

 nscan = nskip*nsave
 skipcount = 0
 isave = 1
 zsave = array(NA, c(nsave, m,n) )

 thetasave = matrix(NA, nsave, m )
 phisave = rep(NA, nsave)
 glob.pvalues = rep(NA, nsave)

 replacem=m   ## how many types are modified in each state proposal
              ## using fewer improves acceptance rate

 for(iscan in 1:(nscan+initc))
 {
  # skip the first initc runs
  if (iscan > initc)
  {  
   skipcount = skipcount+1
  }
  
  ######################################################################
  # Update the latent MT/MC counts Z; sample one patient and tweak counts
  # by prior sampling
 
  istar = sample( 1:m, size=replacem, replace=FALSE ) # types to be updated
  jstar = sample( 1:n, size=1)  ## patient to be updated
  zstar = zstate[,jstar]
  yy = mtfreqptMC[,jstar]

  zstar[istar] = rpois( replacem, lambda=muz[istar,jstar] )
  qstar = zstar
  q0 = zstate[,jstar]

  sumy = sum(yy)
  sumqstar = sum(qstar) 
  sumq0 = sum(q0)  

  ok = TRUE
  if( any( (yy>0)&(qstar==0)  ) ){ ok = FALSE }  

  if( ok )
  {
    ii = !( ( yy[istar] == 0 )&(qstar[istar] == 0) )
    jj = !( ( yy[istar] == 0 )&(q0[istar] == 0) )
    kk = ii&jj
    logmh = 0
    if( any(kk) )
     {
      num = lgamma( phi*(qstar[istar])[kk] + (yy[istar])[kk]  ) - 
			lgamma( phi* (qstar[istar])[kk] )
      den = lgamma( phi*(q0[istar])[kk] + (yy[istar])[kk]  ) - 
			lgamma( phi* (q0[istar])[kk] )
      logmh = sum(num) - sum(den)
      
     }
    logmh = logmh + lgamma(phi*sumqstar) - lgamma(sumy+phi*sumqstar) - 
           lgamma(phi*sumq0) + lgamma(sumy+phi*sumq0)

    if( log(runif(1)) <= logmh ) 
     {
      zstate[,jstar] =  zstar
      acrate[1] = acrate[1] + 1
     }
  }
  ###########################################################
  # update phi value (overdispersion parameter)

  ys <- mtfreqptMC
  zs <- zstate
 
  # propose new phi value; random walk
  phistar = abs(runif(1,min=phi-phi.delta,max=phi+phi.delta))

  aa <- phi*zs
  bb <- aa + ys
  aastar <- phistar*zs
  bbstar <- aastar + ys
  ok.a <- (aa > 0)
  ok.b <- (bb > 0)    ## same for both current and proposed states

  tmp.a <- sum( lgamma(aa[ok.a])  ) - sum( lgamma(aastar[ok.a]) )
  tmp.b <- sum( lgamma(bbstar[ok.b])  ) - sum( lgamma(bb[ok.b]) )
 
  n.types = m
  samplesn = c( t(ys) %*% rep(1,n.types) )   
  sumz = c( t(zs) %*% rep(1,n.types) )

  aa <- phi*sumz
  bb <- aa + samplesn
  aastar <- phistar*sumz
  bbstar <- aastar + samplesn
  ok.a <- (aa>0)
  ok.b <- (bb>0)  ## probably always TRUE

  bar.a <- sum( lgamma(aastar[ok.a]) ) - sum( lgamma(aa[ok.a]) )
  bar.b <- sum( lgamma(bb[ok.b]) ) - sum( lgamma(bbstar[ok.b]) )

  logmh <- tmp.a + tmp.b + bar.a + bar.b 
  # include the prior
## ignore the prior...
##  logmh <- logmh + lambda*( phi - phistar )

  if( log(runif(1)) <= logmh ) 
    {
      phi = phistar
      acrate[2] <- acrate[2] + 1
    }
 
  
  ##########################################################
  # update the proportion of V/J-region types (theta) by Gibbs sampling
  ssnew = apply( zstate, 1,sum )
   if (method==1)
   {
    uuu = rgamma( m, shape=ssnew +mtfreqSCDT+1) 
   }
   if (method==2)  ## makes posterior predictive p's
   {
    uuu = rgamma( m, shape=ssnew +mtfreqSCDT+wtfreq+1) 
   }
  
  theta <- uuu/sum(uuu) 
  muz <- outer( theta, nsurvive, "*" )  ## prior mean of hidden state rates
   
  if( skipcount == nskip )
  {
    skipcount = 0
	zsave[isave,,] = zstate
    thetasave[isave,] = theta
    phisave[isave] = phi
	
    if( print.iter ){ print( isave )  }
    isave=isave+1
   }
  }
  acrate <- acrate/(nscan+initc)
  names(acrate) <- c("z state","phi" )
  
  res <- list(zsave=zsave,thetasave=thetasave, phisave=phisave, counts=counts,
	 nsurvive=nsurvive, acrate=acrate) 
  res
 }

